#include "irc.h"
#include "server.h"
#include "edit.h"
#include "crypt.h"
#include "vars.h"
#include "ircaux.h"
#include "lastlog.h"
#include "window.h"
#include "screen.h"
#include "whois.h"
#include "hook.h"
#include "input.h"
#include "ignore.h"
#include "keys.h"
#include "names.h"
#include "alias.h"
#include "history.h"
#include "funny.h"
#include "ctcp.h"
#include "dcc.h"
#include "translat.h"
#include "output.h"
#include "exec.h"
#include "notify.h"
#include "numbers.h"
#include "status.h"
#include "list.h"
#include "pluto.h"
#include "bot.h"
#include "string.h"

#define getrandom(min, max) ((rand() % (int)(((max)+1) - (min))) + (min))
#define CHECK_FOR_OPS		/*
				 * * * * * * * define this if you want to
				 * * * * * * * make sure you are opped before
				 * * * * * * * doing a mode 
				 */

FILE *logptr = NULL;

int split_watch = 0;
int serv_action = 0;
int first_time = 0;
char *last_msg = NULL;
char *last_notice = NULL;

#define SPLIT 1
#define FALSE 0

#ifndef TRUE
#define TRUE ~FALSE
#endif

void newhost(char *channel, char *arg, char *subargs);

extern char MyHostName[80];
extern struct in_addr MyHostAddr;	/*

					 * 
					 * * The local machine address 
					 */


extern Ignore *ignored_nicks;

irc_server *tmplink = NULL, *server_last = NULL;

int isme(char *nick)
{
	return ((my_stricmp(nick, get_server_nickname(from_server)) == 0) ? 1 : 0);
}

void clear_link(irc_server **serv1)
{
	irc_server *temp = *serv1, *hold;

	while (temp != NULL)
	{
		hold = temp->next;
		new_free(&temp->name);
		new_free(&temp->link);
		new_free((char **) &temp);
		temp = hold;
	}
	*serv1 = NULL;
}

void add_server(irc_server **serv1, char *channel, char *arg, int hops)
{
	irc_server *serv2;
	serv2 = (irc_server *) new_malloc(sizeof (irc_server));
	serv2->next = *serv1;
	malloc_strcpy(&serv2->name, channel);
	malloc_strcpy(&serv2->link, arg);
	serv2->hopcount = hops;
	serv2->time = time(NULL);
	*serv1 = serv2;
}

int find_server(irc_server *serv1, char *channel)
{
	irc_server *temp;

	for (temp = serv1; temp; temp = temp->next)
	{
		if (!my_stricmp(temp->name, channel))
			return TRUE;
	}
	return FALSE;
}

void parse_364(char *channel, char *args, char *subargs)
{
	if (!*channel || !*args)
		return;

	add_server(&tmplink, channel, args, atol(subargs));
}

void parse_365(char *channel, char *args, char *subargs)
{
	irc_server *serv1;

	for (serv1 = server_last; serv1; serv1 = serv1->next)
	{
		if (!find_server(tmplink, serv1->name))
		{
			if (!(serv1->status & SPLIT))
				serv1->status = SPLIT;
			if (serv1->count)
				continue;
			if (do_hook(LLOOK_SPLIT_LIST, "%s %s %d", serv1->name, serv1->link, serv1->hopcount))
				put_it("\002%s\002: %s Split from %s %d", version, serv1->name, serv1->link, serv1->hopcount);
			serv1->count++;
		}
		else
		{
			if (serv1->status & SPLIT)
			{
				serv1->status = ~SPLIT;
				if (do_hook(LLOOK_JOIN_LIST, "%s %s %d", serv1->name, serv1->link, serv1->hopcount))
					put_it("\002%s\002: %s Merge %s %d", version, serv1->name, serv1->link, serv1->hopcount);
				serv1->count = 0;
			}
		}
	}
	for (serv1 = tmplink; serv1; serv1 = serv1->next)
	{
		if (!find_server(server_last, serv1->name)) {
			if (first_time == 1)
			{
				if (do_hook(LLOOK_ADDED_LIST, "%s %s %d", serv1->name, serv1->link, serv1->hopcount))
					put_it("\002%s\002: %s Added %s %d", version, serv1->name, serv1->link, serv1->hopcount);
				serv1->count = 0;
			}
			add_server(&server_last, serv1->name, serv1->link, serv1->hopcount);
		}
	}
	first_time = 1;
	clear_link(&tmplink);
}

/*
 * find split servers we hope 
 */
void linklook(char *channel, char *arg, char *subargs)
{
struct server_split *serv = server_last;
int count;
	if (serv == NULL)
	{
		bitchsay("Link Look not active!");
		return;
	}
	
	count = 0;
	while (serv)
	{
		if (serv->status & SPLIT)
		{
			if (!count)
				say("distance		server			upLink");
			if (do_hook(LLOOK_SPLIT_LIST, "%s %s %d", serv->name, serv->link, serv->hopcount))
				say("[%3d] [%28s] <- [%s]", serv->hopcount, serv->name, serv->link);
			count++;
		}
		serv = serv->next;
	}
	if (count)
		bitchsay("There %s %d split servers", (count == 1) ? "is": "are", count);
	else
		bitchsay("No split servers found");
}

void do_dirlastnotice(char *command, char *args, char *subargs)
{
	char *channel = NULL;

	if (args && *args)
		channel = next_arg(args, &args);
	else
		channel = get_channel_by_refnum(0);
	if (is_channel(channel))
	{
		if (last_notice)
			send_text(channel, last_notice, NULL);
	}
	else
		say("No Current Channel for this Window!");
}

void do_dirlastmsg(char *command, char *args, char *subargs)
{
	char *channel = NULL;

	if (args && *args)
		channel = next_arg(args, &args);
	else
		channel = get_channel_by_refnum(0);
	if (is_channel(channel))
	{
		if (last_msg)
			send_text(channel, last_msg, NULL);
		else
			bitchsay("You have no Friends");
	}
	else
		say("No Current Channel for this Window!");
}

/*
 * Wallop   Sends NOTICE to all ops of Current Channel!       
 */
void ChanWallOp(char *command, char *args, char *subargs)
{
	char *channel = NULL;
	char *chops = NULL;
	ChannelList *chan;
	NickList *tmp;
	char	buffer[BIG_BUFFER_SIZE + 1];
	
	if (!args || (args && !*args))
	{
		bitchsay("Usage: WALLMSG msg");
		return;
	}
	if (get_channel_by_refnum(0))
	{
		malloc_strcpy(&channel, get_channel_by_refnum(0));
		chan = lookup_channel(channel, curr_scr_win->server, 0);
		for (tmp = chan->nicks; tmp; tmp = tmp->next)
		{
			if (tmp->chanop == 1)
			{
				if (chops)
					malloc_strcat(&chops, ",");
				malloc_strcat(&chops, tmp->nick);
			}
		}
		if (chops)
		{
			sprintf(buffer, "[\002BWallop\002/\002%s\002] %s", channel, args);
			send_to_server("%s %s :%s", "NOTICE", chops, buffer);
		}
		else
			say("There are no Channel Operators in %s!", channel);
	}
	else
		say("No Current Channel for this Window.");
}

void split_kill(char *path, char **who, char **who_uh, char **server, char **server2, char **dead, char **reason)
{
	char *temp;

	*who = *who_uh = *server = *server2 = *reason = NULL;

	*dead = strstr(path, "for");
	if (!*dead)
		return;
	*dead = strchr(*dead, ' ');
	if (!*dead)
		return;
	(*dead)++;
	temp = strchr(*dead, ' ');
	if (!temp)
		return;
	*(temp - 1) = '\0';
	temp++;
	*who = strchr(temp, ' ');
	if (!*who)
		return;
	(*who)++;
	temp = strchr(*who, ' ');
	if (!temp)
		return;
	*temp = '\0';
	temp += 7;		/*
				 * now pointing at beginning of the Path 
				 */
	*(temp - 2) = '\0';	/*
				 * to fix the below while() loops 
				 */
	if ((*reason = strchr(temp, ' ')))
	{
		**reason = '\0';
		(*reason)++;
	}
	if (!*reason || !**reason)
		*reason = NULL;
	temp = strrchr(temp, '!');
	if (!temp)
		return;
	*temp = '\0';
	temp--;
	while (*temp && (*temp != ' ') && (*temp != '!'))
		temp--;
	*temp = '\0';
	*who_uh = temp + 1;
	if (!*who_uh || !**who_uh)
		*who_uh = NULL;
	temp--;
	while (*temp && (*temp != ' ') && (*temp != '!'))
		temp--;
	*temp = '\0';
	*server = temp + 1;
	if (!*server || !**server)
		*server = NULL;
	else
	{
		if ((temp = strchr(*server, '[')))
		{
			*temp = '\0';
			*server2 = temp + 1;
			if ((temp = strchr(*server2, ']')))
				*temp= '\0';
		}			
	}
}

void utimediff(long int s1, long int us1, long int s2, long int us2, long int *temp, long int *utemp)
{
	*temp = s1 - s2;
	if (us2 > us1)
	{
		(*temp)--;
		us1 += 1000000;
	}
	*utemp = us1 - us2;
}

int logmsg(int rc, char *type, char *from, char *string, char *extra, int flag)
{
	char *timestr;
	time_t t;
	char *filename = NULL;
	char *expand = NULL;

	if (!get_string_var(MSGLOGFILE_VAR) || !get_string_var(CTOOLZ_DIR_VAR))
		return 0;

	if (get_int_var(MSGLOG_VAR) && !logptr)
		flag = 1;

	t = time(NULL);

	switch (flag)
	{
	case 0:
		if (!logptr)
			return 0;
		timestr = update_clock(GET_TIME);
		fprintf(logptr, "[%s] %s%s%s%s%s%s %s\n", timestr,
			rc ? "-> " : "", type, from, extra ? "!" : "",
		extra ? extra : "", (type && (*type == '<')) ? ">" : type,
			string);
		fflush(logptr);
		break;
	case 1:
		if (logptr)
		{
			bitchsay("Already logging messages");
			new_free(&expand);
			return 1;
		}
		malloc_sprintf(&filename, "%s/%s", get_string_var(CTOOLZ_DIR_VAR), get_string_var(MSGLOGFILE_VAR));
		expand = expand_twiddle(filename);
		new_free(&filename);
		if (!(logptr = fopen(expand, "at")))
		{
			set_int_var(MSGLOG_VAR, 0);
			new_free(&expand);
			return 0;
		}
		fprintf(logptr, "MsgLog started at %s", ctime(&t));
		fflush(logptr);
		if (string)
		{
			int i;

			i = logmsg(rc, type, from, string, extra, 0);
			return i;
		}
		bitchsay("Now logging messages to: %s", expand);
		new_free(&expand);
		break;
	case 2:
		if (!logptr)
		{
			bitchsay("Already not logging messages");
			return 1;
		}
		fprintf(logptr, "MsgLog ended at %s", ctime(&t));
		fclose(logptr);
		logptr = NULL;
		break;
	default:
		bitchsay("Bad Flag passed to logmsg");
		return 0;
	}
	return 1;
}

void msglogger(int flag)
{
	char *logfile;

	if (((logfile = get_string_var(MSGLOGFILE_VAR)) == NULL) || !get_string_var(CTOOLZ_DIR_VAR))
	{
		bitchsay("You must set the MSGLOGFILE and CTOOLZ_DIR variables first!");
		set_int_var(MSGLOG_VAR, 0);
		return;
	}
	logmsg(0, NULL, NULL, NULL, NULL, flag ? 1 : 2);
}

void not_on_a_channel()
{
	bitchsay("D'oh! You're not on a channel!");
}

int are_you_opped(char *channel)
{
	return is_chanop(channel, get_server_nickname(from_server));
}

int you_is_opped(char *channel)
{
	return is_chanop(channel, get_server_nickname(from_server));
}

void you_not_opped(char *channel)
{
	say("D'oh! You're not opped on %s", channel);
}

char *terminate(char *string, char *chars)
{
	char *ptr;

	if (!string || !chars)
		return empty_string;
	while (*chars)
	{
		if ((ptr = strrchr(string, *chars)))
			*ptr = '\0';
		chars++;
	}
	return string;
}

int freadln(FILE *stream, char *lin)
{
	char *p;

	do
		p = fgets(lin, 1000, stream);
	while (p && (*lin == '#'));

	if (!p)
		return 0;
	terminate(lin, "\n\r");
	return 1;
}

char *randreason()
{
	int count, min, i;
	FILE *bleah;
	char *filename = NULL;
	char static buffer[BIG_BUFFER_SIZE + 1];
	char *p;

	min = 1;
	count = 0;

	buffer[0] = '\0';
	malloc_sprintf(&filename, "%s/%s.%s", get_string_var(CTOOLZ_DIR_VAR), version, "reasons");
	p = expand_twiddle(filename);
	new_free(&filename);
	if (!(bleah = fopen(p, "r")))
	{
		new_free(&p);
		return NULL;
	}
	new_free(&p);
	while (!feof(bleah))
		if (freadln(bleah, buffer))
			count++;
	fseek(bleah, 0, 0);
	i = getrandom(1, count);
	count = 0;
	while (!feof(bleah) && (count < i))
		if (freadln(bleah, buffer))
			count++;
	fclose(bleah);
	if (*buffer)
		return buffer;
	return NULL;
}

char *get_reason(char *nickname)
{
	char *temp;
static char reason[BIG_BUFFER_SIZE + 1];
	*reason = '\0';	
	temp = randreason();
	if (!temp || !*temp || !nickname )
		return get_string_var(DEFAULT_REASON_VAR);
	if (strstr(temp, "%s"))
		sprintf(reason, temp, nickname);
	else
		sprintf(reason, temp); 
	return reason;
}

char *do_nslookup(char *host)
{
	struct hostent *temp;
	struct in_addr temp1;
	int ip;

	ip = 0;

	if (!host)
		return (char *) NULL;
	if (isdigit(*(host + strlen(host) - 1)))
	{
		ip = 1;
		temp1.s_addr = inet_addr(host);
		temp = gethostbyaddr((char *) &temp1, sizeof (struct in_addr), AF_INET);
	}
	else
	{
		temp = gethostbyname(host);
		if (temp)
			strncpy((caddr_t) & temp1, temp->h_addr, temp->h_length);
		else
			return (char *) NULL;
	}
	if (!temp)
		return (char *) NULL;
	if (ip)
		return (char *) temp->h_name;
	return ((char *) inet_ntoa(temp1));
}

#if 0
void do_reconnect(char *newnick)
{
	say("Reconnecting to server %s", server_list[from_server].name);
	close_server(from_server, "E\002\002rror 31337");
	if (connect_to_server(server_list[from_server].name,
		    server_list[from_server].port, primary_server) != -1)
	{
		if (newnick)
			send_to_server("NICK %s", newnick);
		change_server_channels(primary_server, from_server);
		set_window_server(-1, from_server, 1);
	}
	else
		say("D'oh! Unable to reconnect.  Use /SERVER to connect.");
}
#endif

void do_newuser(char *newnick, char *newusername, char *ircname)
{
	if (newusername)
	{
		if (!newusername)
			newusername = "";
		strmcpy(username, newusername, NAME_LEN);
		if (ircname && *ircname)
			strmcpy(realname, ircname, REALNAME_LEN);
		reconnect_cmd(NULL, newnick, NULL);
	}
}

char *right(string, num)
char *string;
int num;
{
	if (strlen(string) < num)
		return string;
	return (string + strlen(string) - num);
}

char *random_str(int min, int max)
{
	int i, ii;
	static char str[BIG_BUFFER_SIZE+1];

	i = getrandom(min, max);
	for (ii = 0; ii < i; ii++)
		str[ii] = (char) getrandom(97, 122);
	str[ii] = '\0';
	return str;
}

int numchar(char *string, char c)
{
	int num = 0;

	while (*string)
	{
		if (tolower(*string) == tolower(c))
			num++;
		string++;
	}
	return num;
}

char *cluster(char *hostname)
{
	static char result[BIG_BUFFER_SIZE + 1];
	char temphost[BIG_BUFFER_SIZE + 1];
	char *host;

	if (!hostname)
		return NULL;
	host = temphost;
	strcpy(result, "");
	if (strchr(hostname, '@'))
	{
		if (*hostname == '~')
			hostname++;
		strcpy(result, hostname);
		*strchr(result, '@') = '\0';
		if (strlen(result) > 9)
		{
			result[8] = '*';
			result[9] = '\0';
		}
		strcat(result, "@");
		if (!(hostname = strchr(hostname, '@')))
			return NULL;
		hostname++;
	}
	strcpy(host, hostname);

	if (*host && isdigit(*(host + strlen(host) - 1)))
	{
		int i;
		char *tmp;

		tmp = host;
		for (i = 0; i < 3; i++)
			tmp = strchr(tmp, '.') + 1;
		*tmp = '\0';
		strcat(result, host);
		strcat(result, "*");
	}
	else
	{
		char *tmp;
		int num;

		num = 1;
		tmp = right(host, 3);
		if (my_stricmp(tmp, "com") &&
		    my_stricmp(tmp, "edu") &&
		    my_stricmp(tmp, "net") &&
		    (stristr(host, "com") ||
		     stristr(host, "edu")))
			num = 2;
		while (host && *host && (numchar(host, '.') > num))
		{
			if ((host = strchr(host, '.')) != NULL)
				host++;
			else
				return (char *) NULL;
		}
		strcat(result, "*");
		if (my_stricmp(host, temphost))
			strcat(result, ".");
		strcat(result, host);
	}
	return result;
}

void netfinger(char *name)
{
	char *host = NULL;
	char temp[2], tmpstr[BIG_BUFFER_SIZE + 1];
	struct hostent *hp;
	struct servent *sp;
	struct sockaddr_in sin;
	int s, num;
	register FILE *f = NULL;
	register int c;
	register int lastc = 0;

	if (name)
		host = rindex(name, '@');
	if (!name || !host)
	{
		say("Invaild user@host.");
		return;
	}
	*host++ = 0;
	hp = gethostbyname(host);
	if (hp == NULL)
	{
		static struct hostent def;
		static struct in_addr defaddr;
		static char *alist[1];
		static char namebuf[128];

		defaddr.s_addr = inet_addr(host);
		if (defaddr.s_addr == -1)
		{
			say("%s is unknown.", host);
			return;
		}
		strcpy(namebuf, host);
		def.h_name = namebuf;
		def.h_addr_list = alist, def.h_addr = (char *) &defaddr;
		def.h_length = sizeof (struct in_addr);

		def.h_addrtype = AF_INET;
		def.h_aliases = 0;
		hp = &def;
	}
	sp = getservbyname("finger", "tcp");
	if (sp == 0)
	{
		say("tcp/finger: unknown service");
		return;
	}
	sin.sin_family = hp->h_addrtype;
	bcopy(hp->h_addr, (char *) &sin.sin_addr, hp->h_length);
	sin.sin_port = sp->s_port;
	s = socket(hp->h_addrtype, SOCK_STREAM, 0);
	if (s < 0)
	{
		say("socket error");
		return;
	}
	put_it("[%s]", hp->h_name);

	if (connect(s,(struct sockaddr *) &sin, sizeof (sin)) < 0)
	{
		perror("connect");
		close(s);
		return;
	}
	write(s, name, strlen(name));
	write(s, "\r\n", 2);
	f = fdopen(s, "r");
	num = 0;
	strcpy(tmpstr, "");
	temp[1] = '\0';
	while ((c = getc(f)) != EOF)
	{
		switch (c)
		{
		case 0210:
		case 0211:
		case 0212:
		case 0214:
			c -= 0200;
			break;
		case 0215:
			c = '\n';
			break;
		}
		num++;
		lastc = c;
		if (isprint(c) || isspace(c))
		{
			if ((c == '\n') || (c == '\r') || (num >= 250))
			{
				num = 0;
				if (*tmpstr)
					put_it("%s", tmpstr);
				strcpy(tmpstr, "");
			}
			else
			{
				temp[0] = c;
				strcat(tmpstr, temp);
			}

		}
		else
			break;
/*
 * putchar(c ^ 100);
 */
	}
	if (lastc != '\n')
		put_it("%s", tmpstr);
	(void) fclose(f);
	return;
}

#if 0
void uh_notify(WhoisStuff *stuff, char *nick, char *args)
{
	if (!stuff || !stuff->nick || !nick || !strcmp(stuff->user, "<UNKNOWN>") || my_stricmp(stuff->nick, nick))
		return;
	put_it("\002*** Signon by %s (%s@%s) detected\002", stuff->nick,
	       stuff->user, stuff->host);
}
#endif

void userhost_nsl(WhoisStuff *stuff, char *nick, char *args)
{
	char *nsl;

	if (!stuff || !stuff->nick || !nick || !strcmp(stuff->user, "<UNKNOWN>") || my_stricmp(stuff->nick, nick))
	{
		say("No information for %s", nick);
		return;
	}
	nsl = do_nslookup(stuff->host);
	say("%s is %s", stuff->host, nsl ? nsl : "unknown.");
}

void userhost_ignore(stuff, nick, args)
WhoisStuff *stuff;
char *nick;
char *args;
{
	char *temp;
	int old_window_display;
	char ignorebuf[BIG_BUFFER_SIZE+1];
	Ignore **igptr;

	if (!stuff || !stuff->nick || !nick || !strcmp(stuff->user, "<UNKNOWN>") || my_stricmp(stuff->nick, nick))
	{
		say("No match for user %s", nick);
		return;
	}
	temp = stuff->user;
	if (*temp == '~')
		temp++;
	if (*temp == '#')
		temp++;
	if (!strcmp(args, "+HOST"))
		sprintf(ignorebuf, "*@%s ALL -CRAP -PUBLIC", cluster(stuff->host));
	else if (!strcmp(args, "-HOST"))
		sprintf(ignorebuf, "*@%s ALL -CRAP -PUBLIC", cluster(stuff->host));
	else if (!strcmp(args, "+USER"))
		sprintf(ignorebuf, "*%s@%s ALL -CRAP -PUBLIC", temp,
			cluster(stuff->host));
	else if (!strcmp(args, "-USER"))
	{
		sprintf(ignorebuf, "*!*%s@%s", temp, stuff->host);
		igptr = &ignored_nicks;
		while (*igptr != NULL)
		{
			if (wild_match(ignorebuf, (*igptr)->nick) ||
			    wild_match(nick, (*igptr)->nick))
			{
				sprintf(ignorebuf, "*!*%s@%s NONE", temp, stuff->host);
				old_window_display = window_display;
				window_display = 0;
				ignore(NULL, ignorebuf, ignorebuf);
				window_display = old_window_display;
				say("Unignored %s!%s@%s", nick, stuff->user, stuff->host);
			}
			if (*igptr)
				igptr = &(*igptr)->next;
		}
		return;
	}
	old_window_display = window_display;
	window_display = 0;
	ignore(NULL, ignorebuf, ignorebuf);
	window_display = old_window_display;
	if (strchr(ignorebuf, ' '))
		*strchr(ignorebuf, ' ') = '\0';
	if (*args == '+')
		say("Now ignoring ALL except CRAP and PUBLIC from %s", ignorebuf);
	else
		say("No longer ignoring anything from %s", ignorebuf);
	return;
}

void nkill(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char *people;

	while ((people = new_next_arg(args, &args)))
		send_to_server("NICK %s %s", people,
			       get_string_var(HACKPASS_VAR));
}

void hmode(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char *to, *themsg;

	if (!(to = new_next_arg(args, &args)))
		to = "*";
	themsg = args;
	if (!is_channel(to) && strcmp(to, "*"))
	{
		themsg = to;
		if (args && *args)
			*(themsg + strlen(themsg)) = ' ';
		to = "*";
	}
	if (!strcmp(to, "*"))
		if ((to = get_channel_by_refnum(0)) == NULL)
		{
			not_on_a_channel();
			return;
		}
	send_to_server("MODE %s,%s %s", to, get_string_var(HACKPASS_VAR),
		       themsg);
}

void reset(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char e;

	e = (char) 27;
	printf("%c[2J%cc%c(B%c%c%c[2J", e, e, e, e, e, e);
	refresh_screen(0, NULL);
}

void do_flood(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char buffer[BIG_BUFFER_SIZE + 1];
	char *to, *what, *numb;
	int num, repeat;

	what = NULL;
	numb = NULL;
	if (!(to = new_next_arg(args, &args)))
	{
		say("D'oh!  No object specified for the flood.");
		return;
	}
	else if (!(what = new_next_arg(args, &args)));
	else
		numb = new_next_arg(args, &args);
	if (!strcmp(to, "*"))
		if ((to = get_channel_by_refnum(0)) == NULL)
		{
			not_on_a_channel();
			return;
		}
	what = what ? what : "\007";
	repeat = atoi(numb ? numb : "2");
	if (repeat < 0)
		repeat = 2;

	num = 256 / (strlen(what) + 2);
	strcpy(buffer, "");
	while (num--)
	{
		strcat(buffer, "\001");
		strcat(buffer, what);
		strcat(buffer, "\001");
	}
	while (repeat--)
	{
		say("Flooded the hell out of %s with %s", to, what);
		send_to_server("PRIVMSG %s :%s", to, buffer);
	}
}

extern char *channel_key _((char *));

void cycle(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char *to;

	if (!(to = new_next_arg(args, &args)))
		to = "*";
	if (!is_channel(to) && strcmp(to, "*"))
	{
		say("Umm...I need a channel name");
		return;
	}
	if (!strcmp(to, "*"))
		if ((to = get_channel_by_refnum(0)) == NULL)
		{
			not_on_a_channel();
			return;
		}
	send_to_server("PART %s", to);
	send_to_server("JOIN %s %s", to, channel_key(to));
}

void tochanops(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char buffer[BIG_BUFFER_SIZE + 1];
	char opsmsg[256];
	char *to, *themsg, *wallop_msg;
	ChannelList *Chan;
	NickList *Nick;

	themsg = NULL;
	if (!(to = new_next_arg(args, &args)))
		to = "*";
	if (!is_channel(to) && strcmp(to, "*"))
	{
		themsg = to;
		if (args && *args)
			*(themsg + strlen(themsg)) = ' ';
		to = "*";
	}
	if (!strcmp(to, "*"))
		if ((to = get_channel_by_refnum(0)) == NULL)
		{
			not_on_a_channel();
			return;
		}
	if ((!themsg && !(themsg = args)) || !*themsg)
	{
		say("No message specified");
		return;
	}
	Chan = lookup_channel(to, from_server, 0);
	if (!Chan)
	{
		say("Error looking up nicks for %s", to);
		return;
	}
	*buffer = '\0';
	for (Nick = Chan->nicks; Nick; Nick = Nick->next)
	{
		if (Nick->chanop)
		{	
			strcat(buffer, ",");
			strcat(buffer, Nick->nick);
		}
	}
	if ((*themsg == '-') && (*(themsg + 1) == ' '))
	{
		char *ptr;

		ptr = index(themsg, ' ');
		send_to_server("PRIVMSG %s :%s", buffer + 1, ptr);
		say("Sent to ops on %s: %s", to, ptr);
	}
	else
	{
		wallop_msg = get_string_var(WALLOP_MSG_VAR);
		if (wallop_msg)
			sprintf(opsmsg, wallop_msg, to, to, to, to, to);
		else
			*opsmsg = '\0';
		send_to_server("NOTICE %s :%s %s",
			       buffer + 1, opsmsg, themsg);
		say("Sent to ops on %s: %s", to, themsg);
	}
}

void bomb(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char *to, *tag;

	if ((to = new_next_arg(args, &args)))
	{
		if (!strcmp(to, "*"))
			if ((to = get_channel_by_refnum(0)) == NULL)
			{
				not_on_a_channel();
				return;
			}
		if (!(tag = new_next_arg(args, &args)))
			tag = "UTC";
		say("Bombed %s with %s %s", to, tag, (args && *args) ? args : "");
		if (args && *args)
			send_to_server("NOTICE %s :\001%s %s\001", to, tag, args);
		else
			send_to_server("NOTICE %s :\001%s\001", to, tag);
	}
	else
		say("No object specified to bomb");
}

void nslookup(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char *host, *hostname;

	if ((host = new_next_arg(args, &args)))
	{
		bitchsay("Checking tables...");
		if (!strchr(host, '.'))
			add_to_userhost_queue(host, userhost_nsl, "%s", host);
		else
		{
			hostname = do_nslookup(host);
			say("%s is %s", host, hostname ? hostname : "unknown.");
		}
	}
	else
		bitchsay("/nslookup [[nick][host]]");
}

/*
 * newhost() taken from Hendrix's +multi version...
 */
void newhost(char *command, char *args, char *subargs)
{
	char *newhname;
	struct hostent *hp;

	if ((newhname = new_next_arg(args, &args)))
	{
		strmcpy(MyHostName, newhname, sizeof (MyHostName));
		if ((hp = gethostbyname(MyHostName)) != NULL)
		{
			bcopy(hp->h_addr, (char *) &MyHostAddr, sizeof (MyHostAddr));
/*
 * local_ip_address.s_addr = ntohl(MyHostAddr.s_addr); CDE 
 */
		}
		reconnect_cmd(NULL, NULL, NULL);
	}
	else
		say("You must specify a new hostname");
}


void newnick(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char *newnick, *newusername;

	if ((newnick = new_next_arg(args, &args)) &&
	    (newusername = new_next_arg(args, &args)))
		do_newuser(newnick, newusername, args);
	else
		say("You must specify a nick and username");
}


void newuser(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char *newusername;

	if ((newusername = new_next_arg(args, &args)))
		do_newuser(NULL, newusername, args);
	else
		say("You must specify a username.");
}

void start_finger(stuff, nick, args)
WhoisStuff *stuff;
char *nick, *args;
{
	static char finger_userhost[80] = "\0";

	if (!stuff || !stuff->nick || !nick || !strcmp(stuff->user, "<UNKNOWN>") || my_stricmp(stuff->nick, nick))
	{
		say("No information for %s", nick);
		return;
	}

	sprintf(finger_userhost, "%s@%s", stuff->user, stuff->host);
	if (finger_userhost[0] == '~' || finger_userhost[0] == '#')
	{
		say("Launching finger for %s (%s)", nick, &finger_userhost[1]);
		netfinger(&finger_userhost[1]);
		return;
	}
	say("Launching finger for %s (%s)", nick, finger_userhost);
	netfinger(finger_userhost);
}

void finger(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	char *userhost;

	if ((userhost = new_next_arg(args, &args)))
	{
		if (!strchr(userhost, '@'))
		{
			add_to_userhost_queue(userhost, start_finger, userhost);
			return;
		}
		netfinger(userhost);
	}
	else
		bitchsay("Please specify a user@host or nick to finger.");
}

void do_ig(command, args, subargs)
char *command, *args, *subargs;
{
	char *nickname;
	static char ignore_type[6];

	if (!args || !*args)
	{
		say("Specify a nick to ignore!");
		return;
	}

	while ((nickname = next_arg(args, &args)))
	{
		if (!nickname || !*nickname)
			return;
		if (!my_stricmp(command, "IGH"))
			strcpy(ignore_type, "+HOST");
		else if (!my_stricmp(command, "IG"))
			strcpy(ignore_type, "+USER");
		if (!my_stricmp(command, "UNIGH"))
			strcpy(ignore_type, "-HOST");
		else if (!my_stricmp(command, "UNIG"))
			strcpy(ignore_type, "-USER");
		add_to_userhost_queue(nickname, userhost_ignore, ignore_type);
	}
}



void users(command, args, subargs)
char *command;
char *args;
char *subargs;
{
	ChannelList *chan;
	NickList *nicks;
	char *to, *spec, *rest, *temp1;
	char modebuf[BIG_BUFFER_SIZE + 1];
	char msgbuf[BIG_BUFFER_SIZE +1];
	int count, ops, msg;
	int hook = 0;

	rest = NULL;
	spec = NULL;
	temp1 = NULL;
	strcpy(msgbuf, "");
	if (!(to = next_arg(args, &args)))
		to = "*";
	if (!is_channel(to) && my_stricmp(to, "*"))
	{
		spec = to;
		to = "*";
	}
	if (!my_stricmp(to, "*"))
		if ((to = get_channel_by_refnum(0)) == NULL)
		{
			not_on_a_channel();
			return;
		}
	if (!spec && !(spec = next_arg(args, &args)))
		spec = "*!*@*";
	if (*spec == '-')
	{
		temp1 = spec;
		spec = "*!*@*";
	}
	else
		temp1 = next_arg(args, &args);
	ops = 0;
	msg = 0;
	if (temp1 && !my_strnicmp(temp1, "-ops", strlen(temp1)))
		ops = 1;
	if (temp1 && !my_strnicmp(temp1, "-nonops", strlen(temp1)))
		ops = 2;
	if (ops)
		temp1 = new_next_arg(args, &args);

	if (temp1 && !my_strnicmp(temp1, "-msg", strlen(temp1)))
		msg = 1;
	if (temp1 && !my_strnicmp(temp1, "-notice", strlen(temp1)))
		msg = 2;
	if (temp1 && !my_strnicmp(temp1, "-nkill", strlen(temp1)))
		msg = 3;
	if (temp1 && !my_strnicmp(temp1, "-kill", strlen(temp1)))
		msg = 4;
	if (temp1 && !my_strnicmp(temp1, "-kick", strlen(temp1)))
		msg = 5;
	if (temp1 && !my_strnicmp(temp1, "-stats", strlen(temp1)))
		msg = 6;

	if (msg && (msg != 3) && (msg != 4) && (msg != 5) && (msg != 6) && (!args || !*args))
	{
		say("No message given");
		return;
	}

	chan = lookup_channel(to, from_server, 0);
	if (NULL == chan)
	{
		say("Error looking up channel %s", to);
		return;
	}
	nicks = chan->nicks;
	count = 0;
	if (msg == 6 && do_hook(STAT_HEADER_LIST, "%s %s %s %s %s", "Nick", "dops", "kicks","nicks","publics"))
		put_it("Nick        dops  kicks  nicks  publics");
	while (nicks)
	{
		sprintf(modebuf, "%s!%s", nicks->nick,
		      nicks->host ? nicks->host : "<UNKNOWN>@<UNKNOWN>");
		if (wild_match(spec, modebuf) && (!ops ||
					 ((ops == 1) && nicks->chanop) ||
					 ((ops == 2) && !nicks->chanop)))
		{
			if (msg == 3)
				if (!isme(nicks->nick))
					send_to_server("NICK %s %s", nicks->nick,
					   get_string_var(HACKPASS_VAR));
				else
					count--;
			else if (msg == 4)
				if (!isme(nicks->nick))
					send_to_server("KILL %s :%s (%i", nicks->nick,
						       args && *args ? args : get_reason(nicks->nick),
						       count + 1);
				else
					count--;
			else if (msg == 5)
				if (!isme(nicks->nick))
					send_to_server("KICK %s %s :%s", to,
						       nicks->nick, (args && *args) ? args :
						       get_reason(nicks->nick));
				else
					count--;
			else if (msg == 6)
			{
				if (my_stricmp(nicks->nick, get_server_nickname(from_server)))
				{
					if (do_hook(STAT_LIST, "%s %d %d %d %d", nicks->nick, nicks->dopcount, nicks->kickcount, 
						nicks->nickcount, nicks->floodcount))
						put_it("%-10s  %4d   %4d   %4d     %4d", 
							nicks->nick, nicks->dopcount, nicks->kickcount, 
							nicks->nickcount, nicks->floodcount);
				}
			}	
			else if (msg)
			{
				if (count)
					strcat(msgbuf, ",");
				strcat(msgbuf, nicks->nick);
			}
			else
			{
				if (!count && do_hook(USERS_HEADER_LIST, "%s %s %s %s %s %s %s", "Level", "aop", "prot", "Channel", "Nick", "+o", "UserHost"))
					put_it("UL:     Channel      Nick      +o    UserHost");
				
				if ((hook = do_hook(USERS_LIST, "%d %d %d %s %s %s %c",
					nicks->userlist? nicks->userlist->level:0, 
					nicks->userlist?nicks->userlist->aop:0,
					nicks->userlist?nicks->userlist->prot:0, 
					to, nicks->nick, 
					nicks->host, 
					nicks->chanop ? '@' : ' ')))
				{
					put_it("%3d %d %d %-12s %-9s  %c    (%s)", 
						nicks->userlist ? nicks->userlist->level:0,
						nicks->userlist ? nicks->userlist->aop:0, 
						nicks->userlist ? nicks->userlist->prot:0, 
						to, nicks->nick,
						nicks->chanop ? '@' : ' ', 
						nicks->host);
				}
			}
			count++;
		}
		nicks = nicks->next;
	}
	if (!count)
		say("No match of %s on %s", spec, to);
	else if (!msg && !hook)
		bitchsay("End of UserList on %s %d counted", to, count);

	if (msg && (msg != 3) && (msg != 4) && (msg != 5) && (msg != 6) && count)
	{
		put_it("\002-> %c%s%c\002 %s", (msg == 1) ? '*' : '-',
		       msgbuf, (msg == 1) ? '*' : '-', args);
		send_to_server("%s %s :%s", (msg == 1) ? "PRIVMSG" : "NOTICE",
			       msgbuf, args);
	}
}

#if 1
char *convert(char *string)
{
	char *temp1, *temp2;

	temp1 = (char *) stristr(string, get_string_var(PUB_LOGSTR_VAR));
	if (!temp1 || *temp1 == '\0')
		return (string);

	strcpy(convertstring, string);
	temp2 = (char *) stristr(convertstring, get_string_var(PUB_LOGSTR_VAR));
	*temp2++ = '\0';
	strcat(convertstring, "\026");
	strcat(convertstring, temp1);
	while (*temp2 != ' ' && *temp2 != '\0')
		temp2++;
	while (*temp1 != ' ' && *temp1 != '\0')
		temp1++;
	*temp2++ = '\026';
	*temp2 = '\0';
	strcat(convertstring, temp1);
	return (convertstring);
}
#endif
int caps_fucknut(char *crap)
{
	int total = 0, allcaps = 0;

	while (*crap)
	{
		if ((*crap >= 'a' && *crap <= 'z') || (*crap >= 'A' && *crap <= 'Z'))
		{
			total++;
			if (toupper(*crap) == *crap)
				allcaps++;
		}
		crap++;
	}
	if (total)
		if ((float) allcaps / (float) total >= .75 && total > 3)
			return (1);
	return (0);
}
